@@ -1,37 +1,57 @@
 package Net::IMAP::Client::MsgSummary;
 
+use Encode ();
 use Net::IMAP::Client::MsgAddress ();
 
 sub new {
-    my ($class, $data, $part_id) = @_;
+    my ($class, $data, $part_id, $has_headers) = @_;
+
     bless my $self = {}, $class;
+
     if ($part_id) {
         $self->{part_id} = $part_id;
     }
-    if ($data->{BODY}) {
-        $self->_parse_body($data->{BODY});
+
+    my $tmp = $data->{BODY};
+    if ($tmp) {
+        $self->_parse_body($tmp);
     }
-    if ($data->{BODYSTRUCTURE}) {
-        $self->_parse_bodystructure($data->{BODYSTRUCTURE});
+
+    $tmp = $data->{BODYSTRUCTURE};
+    if ($tmp) {
+        $self->_parse_bodystructure($tmp);
     }
-    if ($data->{ENVELOPE}) {
-        $self->_parse_envelope($data->{ENVELOPE});
+
+    $tmp = $data->{ENVELOPE};
+    if ($tmp) {
+        $self->_parse_envelope($tmp);
     }
-    if ($data->{FLAGS}) {
-        $self->{flags} = $data->{FLAGS};
-    }
-    if ($data->{INTERNALDATE}) {
-        $self->{internaldate} = $data->{INTERNALDATE};
-    }
-    if ($data->{'RFC822.SIZE'}) {
-        $self->{rfc822_size} = $data->{'RFC822.SIZE'};
-    }
-    if ($data->{UID}) {
-        $self->{uid} = $data->{UID};
+
+    $self->{flags} = $data->{FLAGS};
+    $self->{internaldate} = $data->{INTERNALDATE};
+    $self->{rfc822_size} = $data->{'RFC822.SIZE'};
+    $self->{uid} = $data->{UID};
+
+    if ($has_headers) {
+        while (my ($key, $val) = each %$data) {
+            if ($key =~ /^body(?:\.peek)?\s*\[\s*header\.fields/i) {
+                $self->{headers} = $val;
+                last;
+            }
+        }
     }
+
     return $self;
 }
 
+sub _decode {
+    my ($str) = @_;
+    if (defined($str)) {
+        eval { $str = Encode::decode('MIME-Header', $str); };
+    }
+    return $str;
+}
+
 sub type { $_[0]->{type} }
 
 sub subtype { $_[0]->{subtype} }
@@ -40,13 +60,22 @@ sub parameters { $_[0]->{parameters} }
 
 sub cid { $_[0]->{cid} }
 
-sub description { $_[0]->{description} }
+sub description { _decode($_[0]->{description}) }
 
 sub transfer_encoding { $_[0]->{transfer_encoding} }
 
 sub encoded_size { $_[0]->{encoded_size} }
 
-sub content_type { "$_[0]->{type}/$_[0]->{subtype}" }
+sub content_type {
+    my ($self) = @_;
+    if ($self->type) {
+        return $self->type . '/' . $self->subtype;
+    }
+    if ($self->multipart) {
+        return 'multipart/' . $self->multipart;
+    }
+    return undef;
+}
 
 sub charset { $_[0]->{parameters}->{charset} }
 
@@ -65,10 +94,12 @@ sub filename {
     unless ($filename) {
         $filename = $_[0]->{parameters}->{name};
     }
-    return $filename;
+    return _decode($filename);
 }
 
-sub multipart { lc $_[0]->{multipart_type} }
+sub name { _decode($_[0]->{parameters}->{name}) }
+
+sub multipart { $_[0]->{multipart_type} }
 
 sub parts { $_[0]->{parts} }
 
@@ -92,7 +123,7 @@ sub language { $_[0]->{language} }
 
 sub date { $_[0]->{date} }
 
-sub subject { $_[0]->{subject} }
+sub subject { _decode($_[0]->{subject}) }
 
 sub from { $_[0]->{from} }
 
@@ -110,6 +141,10 @@ sub in_reply_to { $_[0]->{in_reply_to} }
 
 sub message_id { $_[0]->{message_id} }
 
+sub seq_id { $_[0]->{seq_id} }
+
+sub headers { $_[0]->{headers} }
+
 # utils
 
 sub get_subpart {
@@ -128,6 +163,8 @@ sub has_attachments {
     return $mt && $MT_HAS_ATTACHMENT{$mt} ? 1 : 0;
 }
 
+sub is_message { $_[0]->content_type eq 'message/rfc822' }
+
 sub _parse_body {
     my ($self, $struct) = @_;
 
@@ -139,7 +176,7 @@ sub _parse_body {
           if $part_id;
         my $i = 0;
         @tmp = map { __PACKAGE__->new({ BODY => $_}, $part_id . ++$i) } @tmp;
-        $self->{multipart_type} = $multipart;
+        $self->{multipart_type} = lc $multipart;
         $self->{parts} = \@tmp;
     } else {
         $self->{type} = lc $struct->[0];
@@ -153,7 +190,7 @@ sub _parse_body {
         $self->{transfer_encoding} = $struct->[5];
         $self->{encoded_size} = $struct->[6];
 
-        if ($self->content_type eq 'message/rfc822') {
+        if ($self->is_message) {
             # continue parsing attached message
             $self->_parse_envelope($struct->[7]);
             $self->_parse_body($struct->[8]);
@@ -165,26 +202,30 @@ sub _parse_bodystructure {
     my ($self, $struct) = @_;
 
     if (ref($struct->[0]) eq 'ARRAY') {
-        my @tmp = @$struct;
-
-        pop @tmp; pop @tmp; pop @tmp; # XXX: we don't care about
-                                      # extension data of multipart
-                                      # body yet; does it actually
-                                      # contain anything interesting?
-
-        my $multipart = pop @tmp;
+        my $multipart;
+        my @tmp;
+        foreach (@$struct) {
+            if (ref($_) eq 'ARRAY') {
+                push @tmp, $_;
+            } else {
+                $multipart = $_;
+                last;           # XXX: ignoring the rest (extension data) for now.
+            }
+        }
         my $part_id = $self->{part_id} || '';
         $part_id .= '.'
           if $part_id;
         my $i = 0;
         @tmp = map { __PACKAGE__->new({ BODYSTRUCTURE => $_}, $part_id . ++$i) } @tmp;
-        $self->{multipart_type} = $multipart;
+        $self->{multipart_type} = lc $multipart;
         $self->{parts} = \@tmp;
     } else {
         $self->{type} = lc $struct->[0];
         $self->{subtype} = lc $struct->[1];
-        if ($struct->[2]) {
-            my %tmp = @{$struct->[2]};
+        my $a = $struct->[2];
+        if ($a) {
+            __lc_key_in_array($a);
+            my %tmp = @$a;
             $self->{parameters} = \%tmp;
         }
         $self->{cid} = $struct->[3];
@@ -192,20 +233,24 @@ sub _parse_bodystructure {
         $self->{transfer_encoding} = $struct->[5];
         $self->{encoded_size} = $struct->[6];
 
-        if ($self->content_type eq 'message/rfc822') {
+        if ($self->is_message) {
             # continue parsing attached message
             $self->_parse_envelope($struct->[7]);
             $self->_parse_bodystructure($struct->[8]);
         } elsif ($self->type ne 'text') {
             $self->{md5} = $struct->[7];
-            if ($struct->[8]) {
-                my %tmp = @{$struct->[8]};
-                foreach (values %tmp) {
-                    if (ref($_) eq 'ARRAY') {
-                        my %foo = @{$_};
-                        $_ = \%foo;
+            my $a = $struct->[8];
+            if ($a) {
+                for (my $i = 0; $i < @$a; ++$i) {
+                    $a->[$i] = lc $a->[$i];
+                    ++$i;
+                    if (ref($a->[$i]) eq 'ARRAY') {
+                        __lc_key_in_array($a->[$i]);
+                        my %foo = @{ $a->[$i] };
+                        $a->[$i] = \%foo;
                     }
                 }
+                my %tmp = @$a;
                 $self->{disposition} = \%tmp;
             }
             $self->{language} = $struct->[9];
@@ -213,6 +258,13 @@ sub _parse_bodystructure {
     }
 }
 
+sub __lc_key_in_array {
+    my ($a) = @_;
+    for (my $i = 0; $i < @$a; $i += 2) {
+        $a->[$i] = lc $a->[$i];
+    }
+}
+
 sub _parse_envelope {
     my ($self, $struct)  = @_;
     $self->{date}        = $struct->[0];
@@ -286,8 +338,8 @@ Net::IMAP::Client.
 
     my $summary = shift @summaries;
 
-    print Encode::decode('MIME-Header', $summary->subject);
-    print Encode::decode('MIME-Header', $summary->from->[0]);
+    print $summary->subject;
+    print $summary->from->[0];
 
 =head1 DESCRIPTION
 
@@ -312,7 +364,8 @@ example.
 =head1 API REFERENCE
 
 It contains only accessors that return data as retrieved by the FETCH
-command (i.e. you need to undecode it).
+command.  Parts that may be MIME-word encoded are automatically
+undecoded.
 
 =head2 C<new>  # constructor
 
@@ -360,9 +413,14 @@ Shortcut for C<$self->type . '/' .$self->subtype>.
 
 Returns the charset declaration for this part.
 
+=item C<name>
+
+Returns the name of this part, if found in FETCH response.
+
 =item C<filename>
 
-Returns the file name of this part, if found in FETCH response.
+Returns the file name of this part, if found in FETCH response.  If
+there's no filename it will try C<name>.
 
 =item C<multipart>
 
@@ -409,6 +467,10 @@ Returns the flags of this message.
 
 Returns the UID of this message.
 
+=item C<seq_id>
+
+Returns the sequence number of this message, if it has been retrieved!
+
 =item C<date>
 
 Returns the date of this message (from the Date header).
@@ -459,6 +521,13 @@ return the text/html part of the attached message.
 Tries to determine if this message has attachments.  For now this
 checks if the multipart type is 'mixed', which isn't really accurate.
 
+=item C<is_message>
+
+Returns true if this object represents a message (i.e. has
+content_type eq 'message/rfc822').  Note that it won't return true for
+the toplevel part, but you B<know> that that part represents a
+message. ;-)
+
 =back
 
 =head1 TODO
